import datetime

from lemmings import Task


class TaskSetMeta(type):
    def __new__(cls, className, bases, classDict):
        tasks = TaskSetMeta.inhereted_tasks(bases)
        for name, item in classDict.items():
            if isinstance(item, Task):
                item.set_name(name)
                tasks.append(item)
        classDict["tasks"] = tasks
        return type.__new__(cls, className, bases, classDict)

    @staticmethod
    def inhereted_tasks(bases):
        tasks = []
        for base in bases:
            if hasattr(base, "tasks") and isinstance(base.tasks, list):
                tasks += base.tasks
        return tasks


class TaskSet(object, metaclass=TaskSetMeta):
    tasks = []

    def __init__(self):
        self.is_sequence = self._is_sequence()
        self.duration = None
        self.runs = len(self.tasks)
        self.sleep_time = datetime.timedelta(0)
    def max_runs(self, max_run):
        self.runs = max_run
        return self
    def max_duration(self, duration):
        self.duration = duration
        return self
    def rps(self, rps):
        self.sleep_time = datetime.timedelta(seconds=1/rps)
        return self

    # for override in task sets
    async def setup(self, task_set, run):
        pass
    async def teardown(self, task_set, run):
        pass

    def _is_sequence(self):
        ordered = 0
        for item in self.tasks:
            if item.order > 0:
                ordered += 1
        if ordered > 0:
            if ordered != len(self.tasks):
                print(f"WARNING! Found only {ordered} ordered task(s) from {len(self.tasks)} total")
                print("To run tasks as sequence, all tasks should be ordered")
            else:
                return True
        return False

    @property
    def weight_total(self):
        i = 0
        for t in self.tasks:
            i += t.weight
        return i

    def __str__(self):
        return f"{self.__class__.__name__} [{self.runs}, {self.duration}]: {', '.join(map(str, self.tasks))} [total: {self.weight_total}]"




import httpx
import math

from prometheus_client.registry import REGISTRY

INF = float("inf")
MINUS_INF = float("-inf")

class Influx:
    def __init__(self, registry=REGISTRY):
        self.url = "http://localhost:8086"
        self.db = "mydb"
        self.registry = registry

    def send(self):
        r = httpx.post(f'{self.url}/write?db={self.db}', data={'key': 'value'})
    # 'cpu_load_short,host=server02 value=0.67
    # cpu_load_short,host=server02,region=us-west value=0.55
    # cpu_load_short,direction=in,host=server01,region=us-west value=2.0

    def save(self, *grep):
        output = []
        def is_filter(s):
            for f in grep:
                if f in s:
                    return True
            return False
        for metric in self.registry.collect():
            try:
                om_samples = {}
                for s in metric.samples:
                    if not is_filter(s.name):
                        continue
                    for suffix in ['_created', '_gsum', '_gcount']:
                        if s.name == metric.name + suffix:
                            om_samples.setdefault(suffix, []).append(self.sample_line(s))
                            break
                    else:
                        output.append(self.sample_line(s))
            except Exception as exception:
                exception.args = (exception.args or ('',)) + (metric,)
                raise
            for suffix, lines in sorted(om_samples.items()):
                # output.append('# TYPE {0}{1} gauge\n'.format(metric.name, suffix))
                output.extend(lines)
        return ''.join(output)

    def escape(self, v):
        return v.replace(',', r'\,').replace('=', r'\=').replace(' ', r'\ ')\
                .replace('\\', r'\\').replace('\n', r'\n').replace('"', r'\"')

    def sample_line(self, line):
        items = []
        if line.labels:
            items = ['{0}={1}'.format(k, self.escape(v)) for k, v in sorted(line.labels.items())]
        labels = ','.join(items)
        timestamp = ''
        if line.timestamp is not None:
            timestamp = ' {0:d}'.format(int(float(line.timestamp) * 1000)) # Convert to milliseconds.
        return '{0},{1} value={2} {3}\n'.format(line.name, labels, floatToGoString(line.value), timestamp)

def floatToGoString(d):
    d = float(d)
    if d == INF:
        return '+Inf'
    elif d == MINUS_INF:
        return '-Inf'
    elif math.isnan(d):
        return 'NaN'
    else:
        s = repr(d)
        dot = s.find('.')
        if d > 0 and dot > 6:
            mantissa = '{0}.{1}{2}'.format(s[0], s[1:dot], s[dot + 1:]).rstrip('0.')
            return '{0}e+0{1}'.format(mantissa, dot - 1)
        return s



import os
import shutil

from prometheus_client import CollectorRegistry, multiprocess

from lemmings.utils.influx import Influx


class Prometheus:
    def __init__(self, shared_dir='./prometheus.tmp'):
        self.path = os.environ.setdefault('prometheus_multiproc_dir', shared_dir)
        shutil.rmtree(self.path, ignore_errors=True)
        os.mkdir(self.path)

        self.registry = CollectorRegistry()
        multiprocess.MultiProcessCollector(self.registry)
        self.influx = Influx(self.registry)

    def filter(self, *args):
        self.args = args

    def dump_to_influx(self):
        print(self.influx.save(*self.args))

    def clean(self):
        try:
            shutil.rmtree(self.path)
            print("temporary prometheus dir cleared")
        except BaseException as e:
            print("problem with temporary prometheus dir: ", e)



from lemmings.utils.utils import ignore_exception


class Color:
    e = "\033[0m"  # End of color
    m = "\033[1m"  # bold
    i = "\033[3m"  # cursive
    s = "\033[4m"  # cursive
    black, red, green, yellow, blue, magenta, cyan, white = 0, 1, 2, 3, 4, 5, 6, 7

    @classmethod
    def _(cls, typ, color, bright):
        b = ";1" if bright else ""
        return f"\033[{typ}{color}{b}m"

    @classmethod
    def _foreground(cls, color, bright=False):
        return cls._(3, color, bright)

    @classmethod
    def _light(cls, color, bright=False):
        return cls._(9, color, bright)

    @classmethod
    def _background(cls, color, bright=False):
        return cls._(4, color, bright)

    @classmethod
    @ignore_exception
    def execute(cls, cmd, msg):
        return cmd + str(msg) + cls.e

    @classmethod
    def bold(cls, msg):
        return Color.execute(cls.m, msg)

    @classmethod
    def cursive(cls, msg):
        return Color.execute(cls.i, msg)

    @classmethod
    def underline(cls, msg):
        return Color.execute(cls.s, msg)

    @classmethod
    def mark(cls, color, msg):
        return Color.execute(Color._foreground(color), msg)

    @classmethod
    def back(cls, color, msg):
        return Color.execute(Color._background(color), msg)

    @classmethod
    @ignore_exception
    def print_full(cls, back, color, msg, comment="", prefix=""):
        print(prefix + cls.b[back % len(Color.b)] + cls.c[color % len(Color.c)] + str(msg) + cls.e + str(comment))

    @classmethod
    def print(cls, color, msg, comment="", prefix=""):
        Color.print_full(color, color, msg, comment, prefix)


Color.b = (
    Color._background(Color.black),
    Color._background(Color.black),
    Color._background(Color.red),
    Color._background(Color.yellow),
)
Color.c = (
    # C.fore(C.red),
    # C.back(C.red)+C.light(C.white, True),
    Color._foreground(Color.green),
    Color._light(Color.white),
    Color._foreground(Color.blue),
    Color._foreground(Color.black),
)

if __name__ == '__main__':
    for i in range(len(Color.c)):
        Color.print(i, f" test color#{i}")



import asyncio
import collections
import random
import time
from datetime import timedelta

from lemmings.task import Task
from lemmings.utils.coroutine import duration2sec


class TaskSetRun:
    def __init__(self, worker, task_set):
        self.worker = worker
        self.task_set = task_set
        self.total = len(self.task_set.tasks)

        self.current = -1  # idx in task_set of current task (for generator)
        self.generated = 0  # current count of generated tasks
        for item in self.task_set.tasks:
            item.set_test_plan(task_set)
        self.start_time = time.time()

    @property
    def shared(self):
        return self.worker.shared_state

    def log(self, msg, comment=""):
        self.worker.log(msg, comment)

    async def sleep(self, duration=None, comment=""):
        if not duration:
            duration = timedelta(seconds=random.random())
        if duration.total_seconds() > 3:
            self.log(f"SLEEP FOR {duration}", comment)
        await asyncio.sleep(duration2sec(duration))

    async def in_duration(self, duration, times) -> collections.AsyncIterable:
        start_time = time.time()
        for step in range(times):
            yield step
            elapsed = timedelta(seconds=time.time() - start_time)
            if times - step > 1:
                sleep_time = (duration - elapsed) / (times - step - 1)
                await self.sleep(sleep_time, f"after #{step + 1} step of {times} total in duration {duration}")

    def rps_in_duration(self, duration, rps) -> collections.AsyncIterable:
        return self.in_duration(duration, duration * rps)

    def with_rps(self, rps):
        class RPS:
            def __init__(self, rps, runner):
                self.rps = rps
                self.runner = runner

            def in_duration(self, duration, rps) -> collections.AsyncIterable:
                return self.runner.in_duration(duration, duration * rps)

        return RPS(rps, self)

    @property
    def is_ordered(self):
        return self.task_set.is_sequence

    @property
    def tasks(self):
        return self.task_set.tasks

    @property
    def sorted_tasks(self):
        return sorted(self.tasks, key=lambda t: t.order)

    def setup_task(self):
        return Task(self.task_set.setup, system=True, comment="Setup * ")

    def teardown_task(self):
        return Task(self.task_set.teardown, system=True, comment="Teardown * ")

    def __iter__(self):
        return self

    def __next__(self):
        # timeout by TaskSet's 'max_duration' parameter
        if self.task_set.duration is not None:
            if time.time() - self.start_time >= self.task_set.duration.total_seconds():
                raise StopIteration
        # stop by TaskSet's 'max_run' parameter
        if self.generated > self.task_set.runs:
            raise StopIteration
        if self.is_ordered:
            s = self.ordered()
            if s is None:
                raise StopIteration
            return s
        else:
            return self.weighted_random()

    def _schedule_teardown(self):
        self.in_processing = False
        return Task(self.task_set.teardown, system=True, comment="Teardown * ")

    def weighted_random(self):
        n = random.uniform(0, self.task_set.weight_total)
        for idx, item in enumerate(self.tasks):
            if n < item.weight:
                self.current = idx
                self.generated += 1
                return item
            n = n - item.weight
        return None

    def ordered(self):
        self.current += 1
        if self.current >= self.total:
            return None
        self.generated += 1
        return self.sorted_tasks[self.current]
